/*************************************************************/
/* Authors:                                                  */
/* - Jaime Arias(original python code)                       */
/* - Mauricio Toro(Oz port and partial extension to NTCC)    */
/* - Salim Perchy(total extension to NTCC, object oriented,  */
/*                implication support, documentation and     */
/*                bug fixes)                                 */
/* Project:                                                  */
/*   NTCC Model Checker                                      */
/* Date:                                                     */
/*   16/06/13                                                */
/*************************************************************/
functor
%%% Imported libraries %%%
import
   System
   FD
   Space
%%% Interface libraries %%%
export
   %%%% Class ltlFormula %%%%
   % Description  : Linear Time Logic Formula
   % Functionality: Create, Get, Set, Show, ... 
   % Attributes   : formula, closure, basicFormulas
   % Records      : c, [], <>, o, v, ^, ->, ~
   %                - preposition(constraint:'c'#[var rel value])
   %                  rel: '=:', '>:', '<:', '>=:', '=<:', '\\=:'
   %                  example: 'x=1'#[x '=:' 1]
   %                - unaryFormula(connector:'[]'|'<>'|'o' child:Formula)
   %                - binaryFormula(connector:'v'|'^'|'->' left:Formula right:Formula)
   %%%%%%%%%%%%%%%%%%%%%%%%%%
   ltlFormula:LTLFormula

   %%%% Class ntccProcess %%%%
   % Description  : NTCC process
   % Funcionality : Create, Get, Set, Show, ConstraintEntailment, ...
   % Attributes   : process, variables, range
   % Records      : c = [var rel value]
   %              - tell( c )
   %              - when( c P )
   %              - unless( c P )
   %              - par( [P1 P2 ... ] )
   %              - sum( [P1 P2 ... ] )
   %              - next( P )
   %              - rep( P )
   %              - star( P )
   %%%%%%%%%%%%%%%%%%%%%%%%%%%
   ntccProcess:NTCCProcess
   
   %%%% Class ntccStructure %%%%
   % Description   : NTCC process structure
   % Functionality : Create, Get, Set, ModelCheck, ...
   % Attributes    : nodes, edges
   % Records       : - ntccNode(store:[c1 c2 ...] normal:[P1 P2 ...] temporal:[P1 P2 ...] initial:true|false)
   %                 - ntccStructure(nodes:[N1 N2 ...] edges:[IdNode#[IdNode1 IdNode2 ...] ...])
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   ntccStructure:NTCCStructure

   %%% Class ntccModelChecker %%%%
   % Description   : Model Checker of NTCC processes against a LTL formula
   % Functionality : Create, Get, Set, ModelCheck
   % Attributes    : formula, process
   % Records       : -none-
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   ntccModelChecker:NTCCModelChecker

%%% Definitions %%%
define
   %%% Class Definition (LTLForumula) %%%
   class LTLFormula
      attr
	 formula       % Linear Time Logic formula
	 closure       % Closure of the LTL Formula
	 basicFormulas % Basic formulas of the Closure
	 combinations  % Combinations of negations 

      % Formula initialization
      meth init( Formula <= preposition(constraint:'x=0'#[x '=:' 0]) )
	 formula := Formula
      end

      % Disjunction
      meth disjunction( $ FormulaLeft <= @formula FormulaRight )
	 binaryFormula(connector:'v' left:FormulaLeft right:FormulaRight)
      end

      % Conjuction
      meth conjunction( $ FormulaLeft <= @formula FormulaRight )
	 binaryFormula(connector:'^' left:FormulaLeft right:FormulaRight)
      end

      % Implication
      meth implication( $ FormulaLeft <= @formula FormulaRight )
	 binaryFormula(connector:'->' left:FormulaLeft right:FormulaRight)
      end

      % Negation
      meth negation( $ Formula <= @formula )
	 case Formula
	 of unaryFormula(connector:'~' child:Child) then Child
	 else unaryFormula(connector:'~' child:Formula)
	 end
      end

      % Next
      meth next( $ Formula <= @formula )
	 unaryFormula(connector:'o' child:Formula)
      end

      % Eventually
      meth eventually( $ Formula <= @formula )
	 case Formula
	 of unaryFormula(connector:'<>' child:Child) then Child
	 else unaryFormula(connector:'<>' child:Formula)
	 end
      end

      % Always
      meth always( $ Formula <= @formula )
	 case Formula
	 of unaryFormula(connector:'[]' child:Child) then Child
	 else unaryFormula(connector:'[]' child:Formula)
	 end
      end

      % Getter and Setter
      meth getFormula( $ )
	 @formula
      end

      meth setFormula( Formula )
	 formula := Formula
      end

      % Private method to show formula in readable format
      meth ShowFormula( $ Formula )
	 case Formula
	 of preposition(constraint:_#Constraint) then
	    {Atom.toString {Nth Constraint 1}}#{Nth Constraint 2}#{Int.toString {Nth Constraint 3}}
	 [] unaryFormula(connector:Connector child:Child) then
	    Connector#'('#{self ShowFormula( $ Child )}#')'
	 [] binaryFormula(connector:Connector left:Left right:Right) then
	    '('#{self ShowFormula( $ Left )}#') '#Connector#' ('#{self ShowFormula( $ Right )}#')'
	 else 'Incorrect formula'
	 end
      end
      
      % Show formula in human readable format
      meth showFormula( $ Formula <= @formula )
	 {self ShowFormula( $ Formula )}
      end

      % Private method to remove []'s from formula, i.e. ~[]f = <>~f
       meth RemoveSquare( $ Formula )
       	 case Formula
       	 of unaryFormula(connector:'[]' child:Child) then
	    unaryFormula(connector:'~' child:unaryFormula(connector:'<>' child:{self negation( $ {self RemoveSquare( $ Child )} )}))
	 [] unaryFormula(connector:'~' child:unaryFormula(connector:'[]' child:Child)) then
	    unaryFormula(connector:'<>' child:{self negation( $ {self RemoveSquare( $ Child )} )})
	 [] unaryFormula(connector:Connector child:Child) then
	    unaryFormula(connector:Connector child:{self RemoveSquare( $ Child )})
	 [] binaryFormula(connector:Connector left:Left right:Right) then
	    binaryFormula(connector:Connector left:{self RemoveSquare( $ Left )} right:{self RemoveSquare( $ Right )})
	 else Formula
	 end
       end

       % Remove []'s from formulas
       meth removeSquare( $ Formula <= @formula )
       	 {self RemoveSquare( $ Formula )}
       end

       % Private method to remove v's from formula, i.e. f1 v f2 = ~(~f1 ^ ~f2)
       meth RemoveDisjunction( $ Formula )
       	 case Formula
       	 of binaryFormula(connector:'v' left:Left right:Right) then
	    unaryFormula(connector:'~' child:binaryFormula(connector:'^'
							   left:{self RemoveDisjunction( $ {self negation( $ Left )} )}
							   right:{self RemoveDisjunction( $ {self negation( $ Right )} )}))
	 [] unaryFormula(connector:'~' child:binaryFormula(connector:'v' left:Left right:Right)) then
	    binaryFormula(connector:'^'
			  left:{self RemoveDisjunction( $ {self negation( $ Left )} )}
			  right:{self RemoveDisjunction( $ {self negation( $ Right )} )})
	 [] binaryFormula(connector:Connector left:Left right:Right) then
	    binaryFormula(connector:Connector left:{self RemoveDisjunction( $ Left )} right:{self RemoveDisjunction( $ Right )})
	 [] unaryFormula(connector:Connector child:Child) then
	    unaryFormula(connector:Connector child:{self RemoveDisjunction( $ Child )})
	 else Formula
	 end
       end

       % Remove v's from formulas
       meth removeDisjunction( $ Formula <= @formula )
       	 {self RemoveDisjunction( $ Formula )}
       end

       % Private method to remove ->'s from formulas, i.e. f1 -> f2 = ~f1 v f2
       meth RemoveImplication( $ Formula )
	  case Formula
	  of binaryFormula(connector:'->' left:Left right:Right) then
	     binaryFormula(connector:'v' left:unaryFormula(connector:'~' child:Left) right:Right)
	  [] unaryFormula(connector:'~' child:binaryFormula(connector:'->' left:Left right:Right)) then
	     binaryFormula(connector:'^' left:Left right:unaryFormula(connector:'~' child:Right))
	  [] binaryFormula(connector:Connector left:Left right:Right) then
	     binaryFormula(connector:Connector left:{self RemoveImplication( $ Left )} right:{self RemoveImplication( $ Right )})
	  [] unaryFormula(connector:Connector child:Child) then
	     unaryFormula(connector:Connector child:{self RemoveImplication( $ Child )})
	  else Formula
	  end
       end

       % Remove ->'s form formulas
       meth removeImplication( $ Formula <= @formula )
	  {self RemoveImplication( $ Formula )}
       end

       % Test if two formulas are equal
      meth equalFormulas( $ Formula2 Formula1 <= @formula Process <= nil )
	 case Formula1
	 of preposition(constraint:_#Constraint1) then
	    case Formula2
	    of preposition(constraint:_#Constraint2) then
               %%% Important: Should we use constraintEntailment? (I'm inclined to say we should) %%%
	       Constraint1 == Constraint2 
	    else
	       false
	    end
	 [] unaryFormula(connector:Connector1 child:Child1) then
	    case Formula2
	    of unaryFormula(connector:Connector2 child:Child2) then
	       Connector1 == Connector2 andthen {self equalFormulas( $ Child1 Child2 Process )}
	    else
	       false
	    end
	 [] binaryFormula(connector:Connector1 left:Left1 right:Right1) then
	    case Formula2
	    of binaryFormula(connector:Connector2 left:Left2 right:Right2) then
	       Connector1 == Connector2 andthen
	       if Connector1 == '->' then {self equalFormulas( $ Left1 Left1 )} andthen {self equalFormulas( $ Right1 Right2 )}
	       else
                  %%% Fixed: Binary operators are conmutative, we should also check that. %%%
		  {Bool.'or'
		   ({self equalFormulas( $ Left1 Left2 Process )} andthen {self equalFormulas( $ Right1 Right2 Process )})
		   ({self equalFormulas( $ Left1 Right2 Process )} andthen {self equalFormulas( $ Right1 Left2 Process )})}
	       end
	    else
	       false
	    end
	 end
      end

       % Private method to compute the closure of a formula
       meth CalculateClosure( $ Formula )
	  PositiveFormula NegativeFormula in
	  case Formula
	  of unaryFormula(connector:'~' child:Child) then
	     PositiveFormula = Child
	     NegativeFormula = Formula
	  else
	     PositiveFormula = Formula
	     NegativeFormula = {self negation( $ Formula )}
	  end
	  case PositiveFormula
	  of preposition(constraint:_#_) then [PositiveFormula NegativeFormula]
	  [] binaryFormula(connector:_ left:Left right:Right) then
	     {Append [PositiveFormula NegativeFormula]
	      {Append {self CalculateClosure( $ Left )} {self CalculateClosure( $ Right )}}}
	  [] unaryFormula(connector:Connector child:Child) then
	     case Connector
	     of 'o' then
		{Append [PositiveFormula NegativeFormula unaryFormula(connector:'o' child:{self negation( $ Child )})]
		 {self CalculateClosure( $ Child )} }
	     else
		{Append [PositiveFormula NegativeFormula unaryFormula(connector:'o' child:PositiveFormula)
			      {self negation( $ unaryFormula(connector:'o' child:PositiveFormula) )}
			      unaryFormula(connector:'o' child:NegativeFormula)]
		 {self CalculateClosure( $ Child )}}
	     end
	  end
       end

       % Calculate the closure of a formula
       meth calculateClosure( $ Formula <= @formula )
	  {self CalculateClosure( $ Formula )}
       end

       % Getter & setter of closure
       meth getClosure( $ )
	  @closure
       end

       meth setClosure( Closure )
	  closure := Closure
       end

       % Private method to calculate the basic formulas of a closure
       meth CalculateBasicFormulas( $ Closure )
	  case Closure
	  of Formula|ClosureRest then
	     case Formula
	     of preposition(constraint:_#_) then
		Formula|{self CalculateBasicFormulas( $ ClosureRest )}
	     [] unaryFormula(connector:'o' child:Child) then
		case Child
		of unaryFormula(connector:'~' child:_) then {self CalculateBasicFormulas( $ ClosureRest )}
		else Formula|{self CalculateBasicFormulas( $ ClosureRest )}
		end
	     else {self CalculateBasicFormulas( $ ClosureRest )}
	     end
	  else
	     nil
	  end
       end

       % Calculate the basic formulas of a closure, i.e. the prepositions and next's
       meth calculateBasicFormulas( $ Closure <= @closure  )
	  {self CalculateBasicFormulas( $ Closure )}
       end

       % Get & setter if the basic formulas from the closure
       meth getBasicFormulas( $ )
	  @basicFormulas
       end

       meth setBasicFormulas( BasicFormulas )
	  basicFormulas := BasicFormulas
       end

       % Private method to calculate all combinations
       meth CalculateCombinations( $ Formulas State )
	  fun {CalculateCombination Formulas State}
	     case Formulas
	     of Formula|FormulasRest then
		if State mod 2 == 0 then
		   Formula|{CalculateCombination FormulasRest State div 2}
		else
		   case Formula of unaryFormula(connector:'o' child:Child) then
		      unaryFormula(connector:'o' child:{self negation( $ Child )})|{self negation( $ Formula )}|
		      {CalculateCombination FormulasRest State div 2}
		   else {self negation( $ Formula )}|{CalculateCombination FormulasRest State div 2}
		   end
		end
	     else
		nil
	     end
	  end
       in
	  if State >= 0 then {CalculateCombination Formulas State}|{self CalculateCombinations( $ Formulas State - 1 )}
	  else nil end
       end
       
       % Calculate all combinations
       meth calculateCombinations( $ Formulas )
	  {self CalculateCombinations( $ Formulas {Pow 2 {Length Formulas}} - 1 )}
       end

       % Get & setter of the combinations calculated
       meth getCombinations( $ )
	  @combinations
       end

       meth setCombinations( Combinations )
	  combinations := Combinations
       end

       % Check if a preposition is in a list of formulas
       meth prepositionInCombination( $ Preposition Combination Process )
	  fun {PrepositionConsistentFormula Prep Formula}
	     case Prep
	     of preposition(constraint:_#C1) then
		case Formula
		of preposition(constraint:_#C2) then
		   {Process constraintEntailment( $ C1 C2 false )}
		[] unaryFormula(connector:'~' child:preposition(constraint:_#C2)) then
		   {Process constraintEntailment( $ C1 C2 true )}
		else false
		end
	     [] unaryFormula(connector:'~' child:preposition(constraint:_#C1)) then
		case Formula
		of preposition(constraint:_#C2) then
		   {Process constraintEntailment( $ C1 C2 true )}
		[] unaryFormula(connector:'~' child:preposition(constraint:_#C2)) then
		   {Process constraintEntailment( $ C1 C2 false )}
		else false
		end
	     end
	  end
       in
	  case Combination
	  of Formula|CombinationRest then
	     if {PrepositionConsistentFormula Preposition Formula} then
		true
	     else
		{self prepositionInCombination( $ Preposition CombinationRest Process )}
	     end
	  else false
	  end
       end

       % Check if a formula is consistent with a combination of formulas
       meth formulaConsistentCombination( $ Formula Combination Process )
	  if {self formulaInCombination( $ {self negation( $ Formula )} Combination Process )} then
	     false
	  else
	     case Formula
	     of preposition(constraint:_) then
		{self formulaInCombination( $ Formula Combination Process )} orelse
		{self prepositionInCombination( $ Formula Combination Process )}
	     [] unaryFormula(connector:'~' child:preposition(constraint:_)) then
		{self formulaInCombination( $ Formula Combination Process )} orelse
		{self prepositionInCombination( $ Formula Combination Process )}
	     [] unaryFormula(connector:'o' child:_) then
		{self formulaInCombination( $ Formula Combination Process )}
	     [] unaryFormula(connector:'~' child:unaryFormula(connector:'o' child:_)) then
		{self formulaInCombination( $ Formula Combination Process )}
	     [] unaryFormula(connector:'<>' child:Child) then
		{self formulaInCombination( $ unaryFormula(connector:'o' child:Formula) Combination Process )} orelse
		{self formulaConsistentCombination( $ Child Combination Process )}
	     [] unaryFormula(connector:'[]' child:Child) then
		{self formulaInCombination( $ unaryFormula(connector:'o' child:Formula) Combination Process )} andthen
		{self formulaConsistentCombination( $ Child Combination Process )}
	     [] binaryFormula(connector:'^' left:Left right:Right) then
		{self formulaConsistentCombination( $ Left Combination Process )} andthen
		{self formulaConsistentCombination( $ Right Combination Process )}
	     [] binaryFormula(connector:'v' left:Left right:Right) then
		{self formulaConsistentCombination( $ Left Combination Process )} orelse
		{self formulaConsistentCombination( $ Right Combination Process )}
	     [] binaryFormula(connector:'->' left:Left right:Right) then
		{Not {self formulaConsistentCombination( $ Left Combination Process )}} orelse
		{self formulaConsistentCombination( $ Right Combination Process )}
	     else false
	     end
	  end
       end

       % Check if a formula is in a list of formulas
       meth formulaInCombination( $ Formula Combination Process )
	  case Combination
	  of FormulaC|CombinationRest then
	     if {self equalFormulas( $ Formula FormulaC Process )} then
		true
	     else
		{self formulaInCombination( $ Formula CombinationRest Process )}
	     end
	  else false
	  end
       end

       % Private method to add non basic formulas to the combinations
       meth NonBasicFormulasInCombinations( $ Closure Combinations Process )
	  fun {IsNonBasic Formula}
	     case Formula
	     of unaryFormula(connector:'<>' child:_) then true
	     [] unaryFormula(connector:'[]' child:_) then true
	     [] binaryFormula(connector:_ left:_ right:_) then true
	     else false
	     end
	  end
	  fun {AddNonBasic Formula Combinations}
	     case Combinations
	     of Combination|CombinationsRest then
		if {self formulaConsistentCombination( $ Formula Combination Process )} then
		   (Formula|Combination)|{AddNonBasic Formula CombinationsRest}
		else
		   ({self negation( $ Formula )}|Combination)|{AddNonBasic Formula CombinationsRest}
		end
	     else
		nil
	     end
	  end
       in
	  case Closure
	  of Formula|ClosureRest then
	     if {IsNonBasic Formula} then
		{self NonBasicFormulasInCombinations( $ ClosureRest {AddNonBasic Formula Combinations} Process )}
	     else
		{self NonBasicFormulasInCombinations( $ ClosureRest Combinations Process )}
	     end
	  else
	     Combinations
	  end
       end

       % Add non basic formulas to the combinations
       meth nonBasicFormulasInCombinations( $ Closure <= @closure Combinations <= @combinations Process )
	  {self NonBasicFormulasInCombinations( $ Closure Combinations Process )}
       end

       % IsPrepositionConsistent = prepositionInCombiantion
       % IsInCombination = formulaInCombination
       % IsConsistent = formulaConsistentCombination
       % AddNonBasic = NonBasicFormulasInCombinations
   end

   %%% Class Definition (NTCCProcess) %%%
   class NTCCProcess
      attr
	 process   % NTCC process
	 variables % Variables in the store
	 range     % Range of variables

      % Process initialization
      meth init( Process <= tell([x ':=' 0]) Variables <= [x] Range <= 0#1 )
	 process   := Process
	 range     := Range
	 variables := Variables
      end
      
      % When
      meth when( $ Constraint <= [x ':=' 0] ProcessRight <= @process )
	 when( Constraint ProcessRight )
      end
      
      % Unless
      meth unless( $ Constraint <= [x ':=' 0] ProcessRight <= @process )
	 unless( Constraint ProcessRight )
      end

      % Parallel composition
      meth parallel( $ Process <= @process Processes )
	 par( {Append Process Processes} )
      end

      % Non-determinism
      meth sumation( $ Process <= @process Processes )
	 sum( {Append Process Processes} )
      end

      % Next
      meth next( $ Process <= @process )
	 next( Process )
      end
      
      % Replication
      meth replication( $ Process <= @process )
	 rep( Process )
      end

      % Eventually
      meth eventually( $ Process <= @process )
	 star( Process )
      end

      % Getter and Setter
      meth getProcess( $ )
	 @process
      end

      meth setProcess( Process )
	 process := Process
      end

      % Private method to show a process in human readable form
      meth ShowProcess( $ Process )
	 case Process
	 of [Var Rel Value] then
	    {Atom.toString Var}#Rel#{Int.toString Value}
	 [] tell(C) then
	    'tell('#{self ShowProcess( $ C )}#')'
	 [] when(C P) then
	    'when('#{self ShowProcess( $ C )}#' do '#{self ShowProcess( $ P )}#')'
	 [] unless(C P) then
	    'unless('#{self ShowProcess( $ C )}#' do '#{self ShowProcess( $ P )}#')'
	 [] par( Ps ) then
	    '('#{List.foldR Ps
		 fun {$ P Pr}
		    Op
		    if Pr == '' then Op = '' else Op = ' || ' end
		 in
		    {self ShowProcess( $ P )}#Op#Pr
		 end
		 '' }#')'
	 []  sum( Ps ) then
	    '('#{List.foldR Ps
		 fun {$ P Pr}
		    Op
		    if Pr == '' then Op = '' else Op = ' + ' end
		 in
		    {self ShowProcess( $ P )}#Op#Pr
		 end
		 '' }#')'
	 [] next(P) then
	    'next('#{self ShowProcess( $ P )}#')'
	 [] rep(P) then
	    '!('#{self ShowProcess( $ P )}#')'
	 [] star(P) then
	    '*('#{self ShowProcess( $ P )}#')'
	 else 'Incorrect Process'
	 end
      end

      % Show a process in human readable form
      meth showProcess( $ Process <= @process )
	 {self ShowProcess( $ Process )}
      end

       % Convert NTCC Process constraints to procedures
      meth processToProcedure( $ Process <= @process )
	 case Process
	 of [Var Rel Value] then
	    case Rel
	    of '=:'   then proc {$ Root} Root.Var =:  Value end
	    [] '\\=:' then proc {$ Root} Root.Var \=: Value end
	    [] '>:'   then proc {$ Root} Root.Var >:  Value end
	    [] '<:'   then proc {$ Root} Root.Var <:  Value end
	    [] '>=:'  then proc {$ Root} Root.Var >=: Value end
	    [] '=<:'  then proc {$ Root} Root.Var =<: Value end
	    %else           proc {$ Root} Root.Var =:  Value end
	    end
	 [] tell(C) then
	    tell({self processToProcedure( $ C )})
	 [] when(C P) then
	    when({self processToProcedure( $ C )} {self processToProcedure( $ P )})
	 [] unless(C P) then
	    unless({self processToProcedure( $ C )} {self processToProcedure( $ P )})
	 [] par(Ps) then
	    par({Map Ps fun {$ P} {self processToProcedure( $ P )} end})
	 [] sum(Ps) then
	    sum({Map Ps fun {$ P} {self processToProcedure( $ P )} end})
	 [] next(P) then
	    next({self processToProcedure( $ P )})
	 [] rep(P) then
	    rep({self processToProcedure( $ P )})
	 [] star(P) then
	    star({self processToProcedure( $ P )})
	 %else 'Incorrect Process'
	 end
      end

      % Test whether constraint C1 entails constraint C2 and viceversa
      % meth constraintEntailment( $ C1 C2 NotEntailed <= false )
      % 	 Answer = _
      % 	 proc {Entail ?Ans}
      % 	    Root
      % 	 in
      % 	    Root = {FD.record vars @variables @range}
      % 	    {FD.sumC [1] [Root.{Nth C1 1}] {Nth C1 2} {Nth C1 3}}
      % 	    {FD.reified.sumC [1] [Root.{Nth C2 1}] {Nth C2 2} {Nth C2 3} Ans}
      % 	 end
      % in
      % 	 {Entail Answer}
      % 	 if {IsDet Answer} then
      % 	    if NotEntailed then Answer == 0 else Answer == 1 end
      % 	 else
      % 	    NotEntailed
      % 	 end
      % end

      % Test whether constraint C1 entails constraint C2 and viceversa
      %%% Warning: This function is to be fixed with better constraint entailment %%%
      meth constraintEntailment( $ C1 C2 NotEntailed <= false )
       	 SpaceC1 = {Space.new proc {$ Root} Root = {FD.record vars @variables @range} end}
       	 SpaceC2 = {Space.new proc {$ Root} Root = {FD.record vars @variables @range} end}
       in
       	 if {Nth C1 1} == {Nth C2 1} then
       	    {Space.inject SpaceC1 {self processToProcedure( $ C1 )}}
       	    {Space.inject SpaceC2 {self processToProcedure( $ C2 )}}
	    
       	    {Space.inject SpaceC1 {self processToProcedure( $ C2 )}}
	    {Space.inject SpaceC2 {self processToProcedure( $ C1 )}}
	    
       	    if NotEntailed then
       	       {Space.ask SpaceC1} == failed
       	    else
       	       {Space.ask SpaceC1} == succeeded andthen {Space.ask SpaceC2} == succeeded
       	    end
	 else
            %%% Fixed: The constraints differ on the variables, therefore there is no entailment %%%
	    false
       	 end
      end
   end

   %%% Class Definition (NTCCStructure) %%%
   class NTCCStructure
      attr
	 structure % NTCC structure

      % NTCC Structure initialization
      meth init( Process )
	 structure := {self createStructure( $ Process )}
	 structure := {self reduceStructure( $ )}
	 structure := {self renameStructure( $ )}
      end

      % Getter and Setter
      meth getStructure( $ )
	 @structure
      end

      meth setStructure( Structure )
	 structure := Structure
      end

      meth ShowStructure( Structure )
	 ntccStructure(nodes:Nodes edges:Edges) = Structure
	 Id = {NewCell 1}
	 Process = {New NTCCProcess init}
	 EdgesT
	 case Edges of _#_#_|_ then EdgesT = {List.reverse Edges}
	 else EdgesT = Edges end
      in
	 for N in Nodes do
	    Node
	    case N of IdT#NodeT then
	       Node = NodeT
	       Id := IdT
	    else Node = N end
	 in
	    {System.showInfo '---Node '#@Id#':---'}
	    Id:= @Id + 1
	    {System.printInfo 'Initial:'}
	    {System.print Node.initial}
	    {System.printInfo ' Normal:['}
	    for Normal in Node.normal do
	       {System.printInfo {Process showProcess( $ Normal )}}
	       {System.printInfo ' '}
	    end
	    {System.printInfo '] Temporal:['}
	    for Temporal in Node.temporal do
	       {System.printInfo {Process showProcess( $ Temporal )}}
	       {System.printInfo ' '}
	    end
	    {System.printInfo '] Store:['}
	    for Constraint in Node.store do
	       case Constraint
	       of preposition(constraint:C#_) then
		  {System.printInfo C}
	       [] unaryFormula(connector:'~' child:preposition(constraint:C#_)) then
		  {System.printInfo '~('}
		  {System.printInfo C}
		  {System.printInfo ')'}
	       end
	       {System.printInfo ' '}
	    end
	    {System.showInfo']'}
	 end
	 {System.showInfo '---Edges---'}
	 for Edge in EdgesT do
	    {System.show Edge}
	 end
      end
      
      % Show structure in human readable form
      meth showStructure( Structure <= @structure )
	 {self ShowStructure( Structure )}
      end

      % Method for constructing an NTCCStructure based off an NTCCProcess
      meth createStructure( $ Process )
	 % Return the position of a node in a list of nodes
	 fun {GetPositionOfNode ntccNode(store:Store normal:Normal temporal:Temporal initial:Initial) Nodes Pos} % Pos = 1
	    Formula = {New LTLFormula init}
	 in
	    case Nodes
	    of ntccNode(store:StoreN normal:NormalN temporal:TemporalN initial:_)|NodesRest then
	       if {Length Store}    == {Length StoreN}    andthen
		  {Length Normal}   == {Length NormalN}   andthen
		  {Length Temporal} == {Length TemporalN} andthen 

		  {FoldL Store 
		   fun {$ State Constraint}
		      if {Formula formulaInCombination( $ Constraint StoreN Process )} then State else false end
		   end true} andthen

		  {FoldL Normal
		   fun {$ State Tag} if {Member Tag NormalN} then State else false end
		   end true} andthen
		  
		  {FoldL Temporal
		   fun {$ State Tag} if {Member Tag TemporalN} then State else false end
		   end true} then
		  Pos
	       else
		  {GetPositionOfNode ntccNode(store:Store normal:Normal temporal:Temporal initial:Initial) NodesRest Pos + 1}
	       end
	    else
	       ~1 % The node isn't on the list
	    end
	 end

	 % Extend the structure by one compuational step for each leaf
	 % This function calculates: [NewNodes]#[NewEdges]#[NewLeaves] = [Node]#[IdNode1#IdNode2]#[Id#Node]
	 % Then this is appended to the current Structure
	 fun {ExtendOneStep
	      ntccStructure(nodes:Nodes edges:Edges)#Leaves
	      Id#ntccNode(store:Store normal:Normal temporal:Temporal initial:Initial)}
	    % Calculates the new nodes, leaves and edges of one computational step by this particular leaf
	    fun {ExtendOneStepAux
		 ntccStructure(nodes:Nodes edges:_)#_
		 Id#ntccNode(store:Store normal:Normal temporal:Temporal initial:_)}
	       fun {ConstraintToString Constraint}
		  {Append {Atom.toString {Nth Constraint 1}}
		   {Append {Atom.toString {Nth Constraint 2}} {Int.toString {Nth Constraint 3}}}}
	       end
	    in
               %%% This unit has no more processes left to be executed %%%
	       if Normal == nil then
		  % New node starts with residual processes
		  PossibleNewNode         = ntccNode(store:nil normal:Temporal temporal:nil initial:false)
		  PossibleNewNodePosition = {GetPositionOfNode PossibleNewNode Nodes 1}
	       in
		  % If the node doesn't already exist then it is the new node and new leaf
		  if PossibleNewNodePosition == ~1 then
		     [PossibleNewNode]#[Id#({Length Nodes} + 1)#'e']#[({Length Nodes} + 1)#PossibleNewNode] % External transition
		  else
		     % The possible node exists, no new nodes, no new leaves, edge to the found node
		     nil#[Id#PossibleNewNodePosition#'e']#nil % External transition
		  end
	       %%% The unit has processes than can be advanced %%%
	       else
		  FirstProcess|OtherProcesses = Normal
		  NewNode NewNodeNegative NewNodePositive PossibleNewNodePosition
	       in
		  case FirstProcess
		     
		  %%% Tell %%%
		  of tell(C) then
		     Formula = {New LTLFormula init( preposition(constraint:{ConstraintToString C}#C) )}
		  in
		     if {Formula prepositionInCombination( $ {Formula negation( $ )} Store Process )} then
			nil#nil#nil % Inconsistent store!!!
		     else
			if {Formula formulaInCombination( $ {Formula getFormula( $ )} Store Process )} then
			   NewNode = ntccNode(store:Store normal:OtherProcesses temporal:Temporal initial:false )
			else
			   NewNode = ntccNode(store:{Formula getFormula( $ )}|Store
					      normal:OtherProcesses temporal:Temporal initial:false)
			end
			PossibleNewNodePosition = {GetPositionOfNode NewNode Nodes 1}
			if PossibleNewNodePosition == ~1 then
			   [NewNode]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNode]
			else
			   nil#[Id#PossibleNewNodePosition#'i']#nil
			end
		     end
			
                  %%% When %%%
		  [] when(C P) then
		     Formula = {New LTLFormula init( preposition(constraint:{ConstraintToString C}#C) )}
		  in
		     % Positive Branch
		     if {Formula prepositionInCombination( $ {Formula getFormula( $ )} Store Process )} then
			NewNodePositive = ntccNode(store:Store normal:P|OtherProcesses temporal:Temporal initial:false)
		     else
			NewNodePositive = ntccNode(store:{Formula getFormula( $ )}|Store
						   normal:P|OtherProcesses temporal:Temporal initial:false )
		     end
		     % Negative Branch
		     if {Formula prepositionInCombination( $ {Formula negation( $ )} Store Process )} then
			NewNodeNegative = ntccNode(store:Store normal:OtherProcesses temporal:Temporal initial:false)
		     else
			NewNodeNegative = ntccNode(store:{Formula negation( $ )}|Store
						   normal:OtherProcesses temporal:Temporal initial:false )
		     end
		     % Compute with the new branch
		     % C in store
		     if {Formula prepositionInCombination( $ {Formula getFormula( $ )} Store Process )} then
			% C in store and ~C not in store
			if {Not {Formula prepositionInCombination( $ {Formula negation( $ )} Store Process )}} then
			   PossibleNewNodePosition = {GetPositionOfNode NewNodePositive Nodes 1}
			in
			   if PossibleNewNodePosition == ~1 then
			      [NewNodePositive]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNodePositive]
			   else
			      nil#[Id#PossibleNewNodePosition#'i']#nil
			   end
			end
		     % C not in store
		     else
			% C not in store but ~C in store
			if {Formula prepositionInCombination( $ {Formula negation( $ )} Store Process )} then
			   PossibleNewNodePosition = {GetPositionOfNode NewNodeNegative Nodes 1}
			in
			   if PossibleNewNodePosition == ~1 then
			      [NewNodeNegative]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNodeNegative]
			   else
			      nil#[Id#PossibleNewNodePosition#'i']#nil
			   end
			% C and ~C not in store
			else
			   PossibleNewNodePositivePosition = {GetPositionOfNode NewNodePositive Nodes 1}
			   PossibleNewNodeNegativePosition = {GetPositionOfNode NewNodeNegative Nodes 1}
			in
			   if PossibleNewNodePositivePosition == ~1 then
			      if PossibleNewNodeNegativePosition == ~1 then
				 [NewNodePositive NewNodeNegative]#
				 [Id#({Length Nodes} + 1)#'i' Id#({Length Nodes} + 2)#'i']#
				 [({Length Nodes} + 1)#NewNodePositive ({Length Nodes} + 2)#NewNodeNegative]
			      else
				 [NewNodePositive]#
				 [Id#({Length Nodes} + 1)#'i' Id#PossibleNewNodeNegativePosition#'i']#
				 [({Length Nodes} + 1)#NewNodePositive]
			      end
			   else
			      if PossibleNewNodeNegativePosition == ~1 then
				 [NewNodeNegative]#
				 [Id#({Length Nodes} + 1)#'i' Id#PossibleNewNodePositivePosition#'i']#
				 [({Length Nodes} + 1)#NewNodeNegative]
			      else
				 nil#
				 [Id#PossibleNewNodePositivePosition#'i' Id#PossibleNewNodeNegativePosition#'i']#
				 nil
			      end
			   end
			end
		     end

		  %%% Unless %%%
		  [] unless(C P) then
		     Formula = {New LTLFormula init( preposition(constraint:{ConstraintToString C}#C) )}
		  in
		     % Positive Branch
		     if {Formula prepositionInCombination( $ {Formula getFormula( $ )} Store Process )}  then
			NewNodePositive = ntccNode(store:Store normal:OtherProcesses temporal:Temporal initial:false)
		     else
			NewNodePositive = ntccNode(store:{Formula getFormula( $ )}|Store
						   normal:OtherProcesses temporal:Temporal initial:false)
		     end
		     % Negative Branch
		     if {Formula prepositionInCombination( $ {Formula negation( $ )} Store Process )} then
			NewNodeNegative = ntccNode(store:Store normal:OtherProcesses temporal:P|Temporal initial:false)
		     else
			NewNodeNegative = ntccNode(store:{Formula negation( $ )}|Store
						   normal:OtherProcesses temporal:P|Temporal initial:false)
		     end
                     % Compute with the new branch
		     % C in store
		     if {Formula prepositionInCombination( $ {Formula getFormula( $ )} Store Process )} then
			% C in store and ~C NOT in store
			if {Not {Formula prepositionInCombination( $ {Formula negation( $ )} Store Process )}} then
			   PossibleNewNodePosition = {GetPositionOfNode NewNodePositive Nodes 1}
			in
			   if PossibleNewNodePosition == ~1 then
			      [NewNodePositive]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNodePositive]
			   else
			      nil#[Id#PossibleNewNodePosition#'i']#nil
			   end
			end
		     % C not in store
		     else
			% C NOT in store but ~C in store
			if {Formula prepositionInCombination( $ {Formula negation( $ )} Store Process )} then
			   PossibleNewNodePosition = {GetPositionOfNode NewNodeNegative Nodes 1}
			in
			   if PossibleNewNodePosition == ~1 then
			      [NewNodeNegative]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNodeNegative]
			   else
			      nil#[Id#PossibleNewNodePosition#'i']#nil
			   end
			% C and ~C not in store
			else
			   PossibleNewNodePositivePosition = {GetPositionOfNode NewNodePositive Nodes 1}
			   PossibleNewNodeNegativePosition = {GetPositionOfNode NewNodeNegative Nodes 1}
			in
			   if PossibleNewNodePositivePosition == ~1 then
			      if PossibleNewNodeNegativePosition == ~1 then
				 [NewNodePositive NewNodeNegative]#
				 [Id#({Length Nodes} + 1)#'i' Id#({Length Nodes} + 2)#'i']#
				 [({Length Nodes} + 1)#NewNodePositive ({Length Nodes} + 2)#NewNodeNegative]
			      else
				 [NewNodePositive]#
				 [Id#({Length Nodes} + 1)#'i' Id#PossibleNewNodeNegativePosition#'i']#
				 [({Length Nodes} + 1)#NewNodePositive]
			      end
			   else
			      if PossibleNewNodeNegativePosition == ~1 then
				 [NewNodeNegative]#
				 [Id#({Length Nodes} + 1)#'i' Id#PossibleNewNodePositivePosition#'i']#
				 [({Length Nodes} + 1)#NewNodeNegative]
			      else
				 nil#
				 [Id#PossibleNewNodePositivePosition#'i' Id#PossibleNewNodeNegativePosition#'i']#
				 nil
			      end
			   end
			end
		     end

		  %%% Next %%%
		  [] next(P) then
		     NewNode = ntccNode(store:Store normal:OtherProcesses temporal:P|Temporal initial:false)
		     PossibleNewNodePosition = {GetPositionOfNode NewNode Nodes 1}
		     if PossibleNewNodePosition == ~1 then
			[NewNode]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNode]
		     else
			nil#[Id#PossibleNewNodePosition#'i']#nil
		     end

                  %%% Replication %%%
		  [] rep(P) then
		     %%% Fixed : Temporals are NOT discarded when there is a replication %%%
		     NewNode = ntccNode(store:Store normal:P|OtherProcesses temporal:rep(P)|Temporal initial:false )
		     PossibleNewNodePosition = {GetPositionOfNode NewNode Nodes 1}
		     if PossibleNewNodePosition == ~1 then
			[NewNode]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNode]
		     else
			nil#[Id#PossibleNewNodePosition#'i']#nil
		     end

                  %%% Asynchrony %%%
		  [] star(P) then
		     PossibleNewNodePositivePosition PossibleNewNodeNegativePosition
		  in
		     NewNodePositive = ntccNode(store:Store normal:P|OtherProcesses temporal:Temporal initial:false )
		     NewNodeNegative = ntccNode(store:Store normal:OtherProcesses temporal:star(P)|Temporal initial:false )
		     PossibleNewNodePositivePosition = {GetPositionOfNode NewNodePositive Nodes 1}
		     PossibleNewNodeNegativePosition = {GetPositionOfNode NewNodeNegative Nodes 1}
		     if PossibleNewNodePositivePosition == ~1 then
			if PossibleNewNodeNegativePosition == ~1 then
			   [NewNodePositive NewNodeNegative]#
			   [Id#({Length Nodes} + 1)#'i' Id#({Length Nodes} + 2)#'i']#
			   [({Length Nodes} + 1)#NewNodePositive ({Length Nodes} + 2)#NewNodeNegative]
			else
			   [NewNodePositive]#
			   [Id#({Length Nodes} + 1)#'i' Id#PossibleNewNodeNegativePosition#'i']#
			   [({Length Nodes} + 1)#NewNodePositive]
			end
		     else
			if PossibleNewNodeNegativePosition == ~1 then
			   [NewNodeNegative]#
			   [Id#({Length Nodes} + 1)#'i' Id#PossibleNewNodePositivePosition#'i']#
			   [({Length Nodes} + 1)#NewNodeNegative]
			else
			   nil#
			   [Id#PossibleNewNodePositivePosition#'i' Id#PossibleNewNodeNegativePosition#'i']#
			   nil
			end
		     end

		  %%% Parallel composition %%%
		  [] par(Ps) then
		     NewNode = ntccNode(store:Store normal:{Append Ps OtherProcesses} temporal:Temporal initial:false)
		     PossibleNewNodePosition = {GetPositionOfNode NewNode Nodes 1}
		     if PossibleNewNodePosition == ~1 then
			[NewNode]#[Id#({Length Nodes} + 1)#'i']#[({Length Nodes} + 1)#NewNode]
		     else
			nil#[Id#PossibleNewNodePosition#'i']#nil
		     end

		  %%% Non-determinism %%%
		  [] sum(Ps) then
		     I = {NewCell 1}
		     in
		     {FoldL Ps
		      fun {$ NodesR#Transitions#Leaves P}
			 NewNode       = ntccNode(store:Store normal:P|OtherProcesses temporal:Temporal initial:false )
			 NewTransition = Id#({Length Nodes} + @I)#'i'
			 NewLeaf       = ({Length Nodes} + @I)#NewNode
		      in
			 I := @I + 1
			 {Append NodesR [NewNode]}#{Append Transitions [NewTransition]}#{Append Leaves [NewLeaf]}
		      end
		      nil#nil#nil}

		  %%% Important: Locality is NOT yet implemented %%%
		  else
		     nil#nil#nil
		  end
	       end
	    end
	    NewNodes#NewEdges#NewLeaves = {ExtendOneStepAux
					   ntccStructure(nodes:Nodes edges:Edges)#Leaves
					   Id#ntccNode(store:Store normal:Normal temporal:Temporal initial:Initial)}
	 in
	    ntccStructure(nodes:{Append Nodes NewNodes} edges:{Append NewEdges Edges})#{Append NewLeaves Leaves}
	 end

	 % Extend the structure by all the unresolved leaves
	 fun {ExtendStructure Structure#Leaves}
	    if Leaves == nil then Structure#nil else {ExtendStructure {FoldL Leaves ExtendOneStep Structure#nil}} end
	 end
	 
	 InitialNode       = ntccNode(store:nil normal:[{Process getProcess( $ )}] temporal:nil initial:true)
	 InitialStructure  = ntccStructure(nodes:[InitialNode] edges:nil)
	 InitialLeaves     = [1#InitialNode]
	 ExtendedStructure
      in
	 ExtendedStructure#nil = {ExtendStructure InitialStructure#InitialLeaves}
	 ExtendedStructure
      end

      % Reduce structure to its minimal form(external transitions only)
      meth reduceStructure( $ ExtendedStructure <= @structure )
	 % Calculate external transitions from a node, transition of the form Node1#Node2
	 fun {FindSuccessors ntccStructure(nodes:_ edges:Edges) NodeId} % NodeId <= 1
	    % Find the internal and external transitions from NodeId
	    InternalTransitions#ExternalTransitions =
	    {FoldL Edges
	     fun {$ Internals#Externals Node1#Node2#Transition}
		if Node1 == NodeId then
		   if Transition == 'i' then
		      ((Node1#Node2)|Internals)#Externals
		   elseif Transition == 'e' then
		      Internals#((Node1#Node2)|Externals)
		   end
		else
		   Internals#Externals
		end
	     end
	     nil#nil}
	 in
	    if InternalTransitions == nil then ExternalTransitions
	    else {FoldL InternalTransitions
		  fun {$ Transitions _#Node2}
		     {Append {FindSuccessors ntccStructure(nodes:_ edges:Edges) Node2} Transitions}
		  end
		  nil} %%% Possible bug: 'ExternalTransitions' insted of 'nil' %%%
	    end
	 end
	 % Reduce Structure recursive function
	 fun {ReduceStructureAux ntccStructure(nodes:Nodes edges:Edges) Successors TabuList PastSuccessors}
	    % Set Union
	    fun {SetUnion List1 List2}
	       {FoldL List1
		fun {$ Result Elem}
		   if {Member Elem List2} then
		      Result
		   else
		      Elem|Result
		   end
		end List2}
	    end
	    % Create a new node
	    fun {CreateNewNode Id Nodes NodeSuccessors}
	       TheNewNode =  {Nth Nodes Id}
	    in
	       if {FoldL NodeSuccessors
		   fun {$ Rec Node1#_}
		      if Node1 == Id then true else Rec end
		   end
		   false} then
		  ntccNode(store:Store normal:Normal temporal:Temporal initial:_) = TheNewNode
	       in
		  Id#ntccNode(store:Store normal:Normal temporal:Temporal initial:true)
	       else
		  Id#TheNewNode
	       end
	    end
	 in
	    {FoldL Successors
	     fun {$ ntccStructure(nodes:RecNodes edges:RecEdges) Node#Successor}
		SuccessorsNode = {FindSuccessors ntccStructure(nodes:Nodes edges:Edges) Successor}
		ntccStructure(nodes:SucNodes edges:SucEdges) =
		{FoldL SuccessorsNode
		 fun {$ ntccStructure(nodes:RecNodes edges:RecEdges) NewNode#NewSuccessor}
		    if {Member Node#Successor#NewNode#NewSuccessor TabuList} then
		       ntccStructure(nodes:RecNodes edges:RecEdges)
		    else
		       ntccStructure(nodes:SucNodes edges:SucEdges) =
		       {ReduceStructureAux ntccStructure(nodes:Nodes edges:Edges) [NewNode#NewSuccessor]
			(Node#Successor#NewNode#NewSuccessor)|TabuList PastSuccessors}
		    in
		       ntccStructure(nodes:{SetUnion SucNodes RecNodes} edges:{SetUnion SucEdges RecEdges})
		    end
		 end
		 ntccStructure(nodes:nil edges:nil)}
	     in
		ntccStructure(nodes:{SetUnion [{CreateNewNode Node Nodes PastSuccessors}]
				     {SetUnion
				      {SetUnion
				       {Map SuccessorsNode fun {$ Node1#_} {CreateNewNode Node1 Nodes PastSuccessors} end}
				       RecNodes}
				      SucNodes}}
			      edges:{SetUnion
				     {SetUnion
				      {Map SuccessorsNode fun{$ Node1#_} Node#Node1 end}
				      RecEdges}
				     SucEdges})
	     end
	     ntccStructure(nodes:nil edges:nil)}
	 end
	 Successors = {FindSuccessors ExtendedStructure 1}
      in
	 {ReduceStructureAux ExtendedStructure Successors nil Successors}
      end

      % Rename an already reduced structure
      meth renameStructure( $ ReducedStructure <= @structure )
	 ntccStructure(nodes:Nodes edges:Edges) = ReducedStructure
	 %  Create new node numbering
	 fun {CreateRenamingDictionary}
	    RenamingDictionary = {NewDictionary}
	    Counter            = {NewCell 1}
	    NewNodes           = {NewCell nil}
	 in
	    for Id#Node in Nodes do
	       {Dictionary.put RenamingDictionary Id @Counter}
	       Counter  := @Counter + 1
	       NewNodes := {Append @NewNodes [Node]}
	    end
	    @NewNodes#RenamingDictionary
	 end
	 % Properly rename edges
	 fun {RenameEdges RenamingDictionary NumberNodes}
	    % Rename the nodes from the edges with the new numbering
	    NewEdges = {FoldL Edges
			fun {$ RecEdges Node1#Node2}
			   ({Dictionary.get RenamingDictionary Node1}#{Dictionary.get RenamingDictionary Node2})|RecEdges
			end
			nil}
	    % Condense each node with its edges, the result is: [Ni1#[Nf1 Nf2 ...] Ni2#[...]]
	    fun {ChangeEdgesRepresentation NewEdges Id Result}
	       if Id =< 0 then
		  Result
	       else
		  {ChangeEdgesRepresentation NewEdges (Id - 1)
		   (Id#
		    {FoldL NewEdges
		     fun {$ Nodes2 Node1#Node2}
			if Node1 == Id then
			   Node2|Nodes2
			else
			   Nodes2
			end
		     end
		     nil})|Result}
	       end
	    end
	 in
	    % Condense all nodes
	    {ChangeEdgesRepresentation NewEdges NumberNodes nil}
	 end
	 % New node numbering
	 NewNodes#RenamingDictionary = {CreateRenamingDictionary}
	 % New condensed edges list
	 NewEdges = {RenameEdges RenamingDictionary {Length NewNodes}}
      in
	 % New renamed structure
	 ntccStructure(nodes:NewNodes edges:NewEdges)
      end
   end

   %%% Class Definition (NTCCModelChecker) %%%
   class NTCCModelChecker
      attr
	 formula   % LTLFormula
	 process   % NTCCProcess
	 structure % NTCCStructure
	 graph     % Model Checking Graph
	 subgraphs % Model Checking SubGraph
	 scclist   % Strongly Connected Components of the graph

      % Class initialization
      meth init( Formula Process )
	 formula   := Formula
	 process   := Process
	 structure := {New NTCCStructure init( Process )}
      end

      % Getters & Setters
      meth getFormula( $ )
	 @formula
      end

      meth getProcess( $ )
	 @process
      end

      meth setFormula( Formula )
	 formula := Formula
      end

      meth setProcess( Process )
	 process := Process
      end

      % Calculate the model checking graph states
      % The graph is of the form States#Size
      %   State = [Atoms1#Intitial1 ...]
      %   Atoms = [Atom1#Index1 ...]
      meth CalculateGraphStates( $ Structure Combinations Process )
	 % Get the list of formulas from the store not in closure and deduce if store is consistent with a closure
	 % Output: Formulas#Consistent
	 fun {GetStoreFormulasConsistent Store Closure}
	    {FoldL Store
	     fun {$ Formulas#Consistent FormulaStore}
		Formula = {New LTLFormula init( FormulaStore )}
	     in
		if {Formula formulaConsistentCombination( $ FormulaStore Closure Process )} then
		   if {Not {Formula formulaInCombination( $ FormulaStore Closure Process )}} then
		      (FormulaStore|Formulas)#Consistent
		   else
		      Formulas#Consistent
		   end
		else
		   nil#false
		end
	     end
	     nil#true}
	 end
	 % Get the atom indexes and their associated store indexes along with the stores updated
	 fun {GetAtomsStoresCombinations Store Combinations Counter}
	    AtomsForStoreIndexes#_#UpdatedCombinations#FinalCounter =
	    {FoldR Combinations
	     fun {$ Combination AtomsForStoreIndexesR#StoreIndex#UpdatedCombinationsR#CounterR}
		Formulas#Consistent = {GetStoreFormulasConsistent Store Combination}
	     in
		if Consistent then
		   ((CounterR#StoreIndex)|AtomsForStoreIndexesR)#
		   (StoreIndex - 1)#
		   ({Append Formulas Combination}|UpdatedCombinationsR)#
		   (CounterR + 1)
		else
		   AtomsForStoreIndexesR#
		   (StoreIndex - 1)#
		   (Combination|UpdatedCombinationsR)#
		   CounterR
		end
	     end
	     nil#{Length Combinations}#nil#Counter}
	 in
	    AtomsForStoreIndexes#UpdatedCombinations#FinalCounter
	 end
	 
	 ntccStructure(nodes:Nodes edges:_) = Structure

	 % We do it for each node and add the initial status
	 AtomsForStoreIndexesListofLists#UpdatedCombinations#Counter =
	 {FoldL Nodes
	  fun {$ AtomsForStoreIndexesListofListsR#UpdatedCombinationsR#CounterR
	         ntccNode(store:Store normal:_ temporal:_ initial:Initial)}
	     AtomsForStoreIndexes#NewUpdatedCombinations#FinalCounter =
	     {GetAtomsStoresCombinations Store UpdatedCombinationsR CounterR}
	  in
	     ((AtomsForStoreIndexes#Initial)|AtomsForStoreIndexesListofListsR)#
	     NewUpdatedCombinations#
	     FinalCounter
	  end
	  nil#Combinations#1}
      in
	 {FoldL AtomsForStoreIndexesListofLists
	  fun {$ Answer AtomsForStoreIndexes#Initial}
	     {FoldL AtomsForStoreIndexes
	      fun {$ AnswerAtom Atom#Index}
		 ({Nth UpdatedCombinations Index}#Atom#Index)|AnswerAtom
	      end
	      nil}#Initial|Answer
	  end
	  nil}#(Counter - 1)
      end

      % Calculate the model checking graph edges
      meth CalculateGraphEdges( $ Structure Graph Process )
	 % Check if all P's from the next's in a store are in the store of the next state 
	 fun {IsNextState Formulae1 Formulae2}
	    case Formulae1
	    of Formula1|Formulae1Rest then
	       case Formula1
	       of unaryFormula(connector:'o' child:Child) then
		  Formula = {New LTLFormula init( Child )}
	       in
		  if {Formula formulaInCombination( $ Child Formulae2 Process )} then
		     {IsNextState Formulae1Rest Formulae2}
		  else
		     false
		  end
	       else
		  {IsNextState Formulae1Rest Formulae2}
	       end
	    else
	       true
	    end
	 end

	 ntccStructure( nodes:_ edges:EdgesStructure ) = Structure
	 AtomsForEachState#Size                        = Graph

	 Nodes                   = {NewArray 1 Size nil}
	 NodeCounter             = {NewCell 1}
	 InitialNodes            = {NewCell nil}
	 GraphEdges              = {NewArray 1 Size nil}
	 EdgesStructureTraversal = {NewCell EdgesStructure}
      in
	 for State#Initial in AtomsForEachState do
	    CurrentEdgeTuple|RestOfEdges = @EdgesStructureTraversal
	    _#Successors                 = CurrentEdgeTuple
	    EdgesStructureTraversal     := RestOfEdges
	 in
	    for Atom#AtomIndex#AtomClosureIndex in State do
	       % Crate Node
	       Nodes.AtomIndex := AtomIndex#AtomClosureIndex#Initial#Atom#@NodeCounter
	       % Add to initial node if needed
	       if Initial then
		  InitialNodes := AtomIndex#AtomClosureIndex#Atom#@NodeCounter|@InitialNodes
	       end
	       % Verify if there is an edge to be added for each successor
	       for NextNodeIndex in Successors do
		  SucState#_ = {Nth AtomsForEachState NextNodeIndex}
	       in
		  % For each atom in successor
		  for AtomSuc#AtomSucIndex#_ in SucState do
		     % If we can deduce the next, add it to edges!
		     if {IsNextState Atom AtomSuc} then
			GraphEdges.AtomIndex := AtomSucIndex|GraphEdges.AtomIndex
		     end
		  end
	       end
	    end
	    NodeCounter := @NodeCounter + 1
	 end
	 graph(nodes:Nodes initial:@InitialNodes edges:GraphEdges)
      end

      % Calculate the Strongly Connected Components(SCC) of a graph
      % Ouput: [[Node1#Node2#(true|false)#store#Node3 ...]#[Node1] ...]
      % This is a straight copy/paste from Mauricio Toro's code
      meth TarjanSCC( $  Graph )
	 graph(nodes:Ve initial:_ edges:E) = Graph
	 Ve E Index S SCCS Vindex Vlowlink Strongconnect 
	 Index    = {NewCell  0}
	 S        = {NewCell nil}
	 SCCS     = {NewCell  nil}
	 Vindex   = {NewArray 1 {Array.high Ve} ~1}
	 Vlowlink = {NewArray 1 {Array.high Ve} ~1}
      in
         % Compute the strongly connected component where V appears
	 proc {Strongconnect V Ve}
	    Pop   % List pop function
	    Until % Until loop procedure
	    Component           = {NewCell nil}
	    InitialsInComponent = {NewCell nil}
	 in
            % List pop
	    fun {Pop ListCell}
	       X|RestList = @ListCell in
	       ListCell := RestList
	       X
	    end
	    
            % Until loop
	    proc {Until WW VV Ve}
	       _#ClosureIndex#Initial#Atom#State = Ve.WW in
	       Component := WW#ClosureIndex#Initial#Atom#State|@Component
	       if Initial then
		  InitialsInComponent := WW|@InitialsInComponent
	       end
	       if {Not WW == VV} then {Until {Pop S} VV Ve} end
	    end
	    
            % Set the depth index for v to the smallest unused index
	    Vindex.V   := @Index
	    Vlowlink.V := @Index
	    Index      := @Index + 1
	    S          := V|@S
	    
            %Consider successors of V
	    for W in E.V do 
	       if Vindex.W == ~1 then  % 'W.index is undefined'
	          % Successor W has not yet been visited; recurse on it
		  {Strongconnect W Ve}
		  Vlowlink.V := {Min Vlowlink.V Vlowlink.W}
	       else
		  if {Member W @S} then % "W is in S" then
		     Vlowlink.V := {Min Vlowlink.V Vindex.W}
		  end
	       end
	    end
	    
            % If V is a root node, pop the stack and generate an SCC
	    if Vlowlink.V == Vindex.V then
	       Component := nil     % 'starts a new SCC'
	    in
	       {Until {Pop S} V Ve} % until W == V
	       SCCS := @Component#@InitialsInComponent|@SCCS
	    end
	 end
	 
	 for I in 1..{Array.high Ve} do
	    Vcounter#_#_#_#_ = Ve.I in
	    if Vindex.Vcounter == ~1 then {Strongconnect I Ve} end % if V.index is undefined
	 end
	 @SCCS %returns the strongly connected components, a list of lists of vertices
      end

      % Calculate the subgraphs w.r.t a SCC description
      meth CalculateSubGraphs( $ SCCList Graph )
         % Finds if an index is in a list formed by tuples (that contain an index in the first field)
	 fun {TupleMember I Tuples}
	    {FoldL Tuples
	     fun {$ State Index#_#_#_#_}
		if Index == I then true else State end
	     end
	     false}
	 end
         % Gets the intersection between the nodes in a List of Sucessors and the nodes of an SCC
	 fun {GetIntersection SCC SuccessorsIndex Answer}
	    case SCC
	    of (I#_#_#_#_)|SCCRest then
	       if {Member I SuccessorsIndex} then
		  {GetIntersection SCCRest SuccessorsIndex I|Answer}
	       else
		  {GetIntersection SCCRest SuccessorsIndex Answer}
	       end
	    else
	       Answer
	    end
	 end
	 
	 graph(nodes:_ initial:InitialNodes edges:Edges) = Graph
	 ModelCheckingSubgraphs = {NewCell nil} 
      in
	 for SCC#_ in SCCList do
	    if {Length SCC} > 1 then
	       TempNodes           = {NewCell nil}
	       TempEdges           = {NewCell nil}
	       TempInitialNodes    = {NewCell nil}
	       TempNonInitialNodes = {NewCell nil}
	       
	       proc {AddNodesFromIntersection I Initial Atom ClosureIndex State}
		  Intersection = {GetIntersection SCC Edges.I nil} in
		  if {Length Intersection} > 0 then
		     if {Not {TupleMember I @TempNodes}} then 
			TempNodes := I#ClosureIndex#Initial#Atom#State|@TempNodes
			if Initial then
			   TempInitialNodes    := I#ClosureIndex#Atom|@TempInitialNodes
			else
			   TempNonInitialNodes := I#ClosureIndex#Atom|@TempInitialNodes
			end
		     end
		     TempEdges := I#Intersection|@TempEdges % Node#Successors
		  end
	       end
	    in
	       for I#ClosureIndex#Initial#Atom#State in SCC do 
		  {AddNodesFromIntersection I Initial Atom ClosureIndex State}
	       end
	       
	       for I#ClosureIndex#Atom#State in InitialNodes do 
		  {AddNodesFromIntersection I true Atom ClosureIndex State}
	       end
	       ModelCheckingSubgraphs := graph(nodes     :@TempNodes
					       edges     :@TempEdges
					       initial   :@TempInitialNodes
					       nonInitial:@TempNonInitialNodes)|@ModelCheckingSubgraphs
	    else 
	       Element|nil = SCC
	       I#ClosureIndex#Initial#Atom#_ = Element
	       Successors = Edges.I
	    in
	       if {Member I Successors} then
		  ModelCheckingSubgraphs := graph(nodes     :[Element]
						  edges     :[I#[I]]
						  initial   :if Initial then [I#ClosureIndex#Atom] else nil end
						  nonInitial:if Initial then nil else [I#ClosureIndex#Atom] end)|
		  @ModelCheckingSubgraphs
	       end
	    end
	 end
	 @ModelCheckingSubgraphs
      end

      % Check if a graph is self-fulfilling
      % Self-fulfilling means there's a loop in the graph or a path from a formula <>P to a formula P
      meth IsSelfFulfilling( $ Graph Process )
	 graph(nodes:Nodes edges:_ initial:_ nonInitial:NonInitialNodes ) = Graph
      in
	 % If there's just one node, we look for satisfying P from <>P inside it
	 if {Length Nodes} == 1 then
	    (_#_#_#Atoms#_)|nil = Nodes
	 in
	    {FoldL Atoms
	     fun {$ IsPFromStar Formula}
		FormulaLTL = {New LTLFormula init( Formula )}
	     in
		case Formula
		of unaryFormula(connector:'<>' child:Child) then
		   if {FormulaLTL formulaInCombination( $ Child Atoms Process )} then
		      IsPFromStar
		   else
		      false
		   end
		else
		   IsPFromStar
		end
	     end
	     true}
	 % We look for <>P's from all the nodes
	 else
	    % For all the nodes in the subgraph we look for <>'s
	    {FoldL NonInitialNodes
	     fun {$ ApplicationNodes _#_#Atom}
		Answer1 =
		{FoldL Atom
		 % For each atom of each node of the subgraph  we look for <>'s
		 fun {$ ApplicationAtom Formula}
		    FormulaLTL = {New LTLFormula init( Formula )}
		 in
		    case Formula
		    of unaryFormula(connector:'<>' child:Child) then
		       % We found a <>P
		       Answer2 =
		       % We need to look for P in ONLY the atoms from the subgraph nodes
		       {FoldL NonInitialNodes
			fun {$ ApplicationNodes2 _#_#AtomsSCC}
			   if {FormulaLTL formulaInCombination( $ Child AtomsSCC Process )} then
			      true % P exists
			   else
			      ApplicationNodes2 % P does not exist in that particular atom
			   end
			end
			false} % Initially we haven't found P
		    in
		       if Answer2 then
			  ApplicationAtom
		       else
			  false % It is not self-fulfilling because we couldn't find P
		       end
		    else
		       ApplicationAtom
		    end
		 end
		 true} % Initially we haven't found any <>'s
	     in
		if Answer1 then
		   true % Should't this be ApplicationNodes?
		else
		   ApplicationNodes % And this false
		end
	     end
	     false} % Initially is it not self-fulfilling(Is it?)
	 end
      end
      
      % Verify if a LTL formula is fulfilled by an NTCC process
      meth modelCheck( $ Structure <= @structure Formula <= @formula Process <= @process )
	 % Check if the initial nodes from the combinated graph entails a formula
	 fun {InitialNodesEntailFormula SCCGraph TheFormula}
	    graph(nodes:_ edges:_ initial:InitialNodes nonInitial:_) = SCCGraph
	 in
	    {FoldL InitialNodes
	     fun {$ Entails _#_#Atom}
		FormulaLTL = {New LTLFormula init( TheFormula )}
	     in
		if {FormulaLTL formulaInCombination( $ TheFormula Atom Process )} then
		   true
		else
		   Entails
		end
	     end
	     false}
	 end
	 % Devoid the SCCGraph of data and only leave the edges, nodes with Id's, stores from the ntccStructure
	 fun {GetCounterExample SCCGraph Structure}
	    graph(nodes:Nodes edges:Edges initial:_ nonInitial:_) = SCCGraph
	    ntccStructure(nodes:NodesStructure edges:_)           = Structure
	 in
	    graph(nodes:
		     {FoldL Nodes
		      fun {$ NodesR I#_#Initial#_#State}
			 ntccNode(store:Store normal:_ temporal:_ initial:_) = {Nth NodesStructure State}
		      in
			 node(id:I initial:Initial state:State store:Store)|NodesR
		      end
		      nil}
		  edges:Edges)
	 end
      in
	 {Formula setFormula( {Formula removeSquare( $ {Formula removeDisjunction( $ {Formula removeImplication( $ {Formula negation( $ )} )} )} )} )}
	 {Formula setClosure( {Formula calculateClosure( $ )} )}
	 {Formula setBasicFormulas( {Formula calculateBasicFormulas( $ )} )}
	 {Formula setCombinations( {Formula calculateCombinations( $ {Formula getBasicFormulas( $ )} )} )}
	 {Formula setCombinations( {Formula nonBasicFormulasInCombinations( $ {Formula getClosure( $ )}
									    {Formula getCombinations( $ )} Process )} )}
	 graph     := {self CalculateGraphStates( $ {Structure getStructure( $ )} {Formula getCombinations( $ )} Process )}
	 graph     := {self CalculateGraphEdges( $ {Structure getStructure( $ )} @graph Process )}
	 scclist   := {self TarjanSCC( $ @graph )}
	 subgraphs := {self CalculateSubGraphs( $ @scclist @graph )}

	 % Model Checking
	 {FoldL @subgraphs
	  fun {$ Check SCCGraph}
	     if {InitialNodesEntailFormula SCCGraph {Formula getFormula( $ )}} andthen
		{self IsSelfFulfilling( $ SCCGraph Process )} then
		false#{GetCounterExample SCCGraph {Structure getStructure( $ )}}
	     else
		Check
	     end
	  end
	  true#graph(nodes:nil edges:nil)}
      end
   end
   
end
